#!/usr/bin/env python3
"""
A single class (:class:`.InterfaceModel`) is provided it encapsulates the
loading of the interface model from YAML files into a Pythonic representation
"""
import inspect
import logging
import typing
from dataclasses import dataclass
from pathlib import Path

import yaml

from ..errors import InvalidInterfaceModelData
from . import datatypes, entities, enumtypes, literals, macros, structtypes, uniontypes


@dataclass
class Meta:
    """
    Meta data used by the :class:`.emitter.Emitter` to describe attributes of
    the generated C API and bindings/wrappers, such as license, version, and
    documentation
    """

    lic: str  # License string, this is utilized for SPDX
    project: str = "example"
    version: str = "0.0.1"
    prefix: str = "ex"
    brief: str = "yace example"
    full: str = (
        "This description is provided with the default model-meta, to "
        "change it, then make sure you have a meta-label in your model. "
        "meta: {lic: '"
    )


class InterfaceModel(object):
    """
    The :class:`.InterfaceModel` serves as the root of the "tree" describing
    the interface to generate C API and bindings for.
    """

    MAPPING = {
        obj.cls: obj
        for _, obj in (
            inspect.getmembers(entities)
            + inspect.getmembers(macros)
            + inspect.getmembers(datatypes)
            + inspect.getmembers(enumtypes)
            + inspect.getmembers(structtypes)
            + inspect.getmembers(uniontypes)
        )
        if inspect.isclass(obj) and hasattr(obj, "cls")
    }

    def __init__(self):

        self.meta = Meta(lic="UNLICENSED")
        self.entities = []

    @classmethod
    def entity_from_data(cls, data: dict):
        """
        Factory function instantiating entities from data

        The entities are constructed using the mapping in
        InterfaceModel.MAPPING, e.g.::

            entity = InterfaceModel.MAPPING["struct"](data)
        """

        constructor = InterfaceModel.MAPPING.get(data["cls"])
        if not constructor:
            raise InvalidInterfaceModelData(f"No mapping for {data}")

        for attr in ["dtype", "lit"]:
            if attr not in data:
                continue

            adata = data.get(attr)
            if isinstance(adata, dict):
                data[attr] = cls.entity_from_data(adata)
            elif isinstance(adata, str) and attr == "dtype":
                constructor_dtype = InterfaceModel.MAPPING.get(adata)
                if not constructor_dtype:
                    raise InvalidInterfaceModelData(
                        f"No mapping for dtype shorthand: {adata}"
                    )
                data[attr] = constructor_dtype()
            elif isinstance(adata, int) and attr == "lit":
                data[attr] = literals.LiteralDec(val=adata)
            elif isinstance(adata, str) and attr == "lit":
                data[attr] = literals.LiteralString(val=adata)
            else:
                raise InvalidInterfaceModelData(f"Invalid {data}")

        if "members" in data:
            data["members"] = [cls.entity_from_data(mdata) for mdata in data["members"]]
            if None in data["members"]:
                raise InvalidInterfaceModelData(f"Issues with {data}")

        return constructor(**data)

    @classmethod
    def from_data(cls, data: dict):
        """Construct an InterfaceModel using the given 'data'"""

        interface = cls()

        meta_data = data.get("meta", None)
        if meta_data:
            del data["meta"]
            interface.meta = Meta(**meta_data)

        for label, ents in data.items():
            total = len(ents)
            for count, entity_data in enumerate(ents, 1):
                logging.info(f"Processing entity {count} / {total} in {label}")
                entity = cls.entity_from_data(entity_data)
                interface.entities.append(entity)

        return interface

    @classmethod
    def from_path(cls, path: Path):
        """
        Returns a dict composed of the merged content of all yaml-files in the
        given 'path'
        """

        raw = {}
        with path.open() as sfd:
            for key, val in yaml.safe_load(sfd).items():
                if key not in raw:
                    raw[key] = val
                elif key in raw and isinstance(val, list):
                    raw[key] += val
                else:
                    raw[key].update(val)

        return cls.from_data(raw)



#!/usr/bin/env python3
"""
YACE raises exceptions in case of error.

The errors are usually in-formated YAML-files, in-correct InterfaceModel, type
errors in the constructed InterfaceModel, layout violations in the
InterfaceModel.
"""


class InvalidInterfaceModelData(Exception):
    """The data given to entity instantion is not valid"""

    pass



#!/usr/bin/env python3
import argparse
import logging
from pathlib import Path
from pprint import pprint

import yaml

from yace import __version__ as version
from yace.emitter import Emitter
from yace.model.interface import InterfaceModel


def parse_args():
    """Parse command-line arguments"""

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "model",
        nargs="+",
        type=Path,
        help="Path to Yace Interface Model (yim) file",
    )
    parser.add_argument(
        "--targets",
        nargs="+",
        choices=[
            "capi",
            "cpp",
            "ctypes",
            "cython",
            "go",
            "rust",
        ],
        default=["capi"],
        help="Code generator targets",
    )
    parser.add_argument(
        "--output",
        type=Path,
        default=Path("output"),
        help="Path to output directory, for emitted code / artifacts",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {version}",
    )

    return parser.parse_args()


def main():
    """Emit enums, structs, and pretty-printer functions for them"""

    args = parse_args()

    logging.basicConfig(level=logging.DEBUG)

    args.output.mkdir(parents=True, exist_ok=True)

    for model in [InterfaceModel.from_path(model) for model in args.model]:
        emitter = Emitter(model, model.meta, args.targets, args.output)
        emitter.emit_api_def()
        emitter.emit_api_pp()
        emitter.emit_api_test()
        emitter.emit_docgen()



#!/usr/bin/env python3
"""
Two baseclasses are provided:

* :class:`.Entity`
* :class:`.Typedecl`, a base class for type definitions

Entities are the base classes representing the language constructs of C APIs in
the :class:`yace.model.interface.InterfaceModel`. The entities serve as
encapsulated type-checking of YAML-definitions of the language model.
"""
import typing
from dataclasses import dataclass, field


@dataclass
class Typedecl:
    """
    Representation of a type declaration with modifiers for:

    * const
    * static
    * pointer
    * array
    """

    cls: str

    const: bool = False
    static: bool = False
    pointer: int = 0
    array: int = 0

    label: str = "typedecl"


@dataclass
class Void(Typedecl):
    """A void, that is, the type signaling no type"""

    cls: str = "void"
    width: int = 0


@dataclass
class Entity:
    """Base class for entities containing common fields"""

    symbol: str
    doc: typing.Optional[str]



#!/usr/bin/env python3
"""
This module contains bla blaClasses inheriting from this type include data
types with known and fixed storage

* Boolean

  * :class:`.Bool`

* Numerical

  * Unsigned Integers: :class:`.U8`, :class:`.U16`, :class:`.U32`, :class:`.U64`
  * Signed Integers: :class:`.I8`, :class:`.I16`, :class:`.I32`, :class:`.I64`
  * Floating Point: :class:`.F32`, :class:`.F64`

* Textual

  * :class:`.Char`
  * :class:`.String`

"""
import typing
from dataclasses import dataclass, field

from .entities import Typedecl


@dataclass
class Datatype(Typedecl):
    """
    These types should have a known and well-defined storage-width, e.g. an
    unsigned 32-bit integer should consume 4 bytes of memory. Modifiers are
    available:
    """

    width: int = 8


@dataclass
class I8(Typedecl):
    """
    A signed integer, 8 bits wide.

    The C emitter produces::

        int8_t

    for the :class:`I64` entity.
    """

    cls: str = "i8"
    width: int = 8


@dataclass
class I16(Typedecl):
    """
    A signed integer, 16 bits wide.

    The C emitter produces::

        int16_t

    for the :class:`I16` entity.
    """

    cls: str = "i16"
    width: int = 16


@dataclass
class I32(Typedecl):
    """
    A signed integer, 32 bits wide.

    The C emitter produces::

        int32_t

    for the :class:`I32` entity.
    """

    cls: str = "i32"
    width: int = 32


@dataclass
class I64(Typedecl):
    """
    A signed integer, 64 bits wide.

    The C emitter produces::

        int64_t

    for the :class:`I64` entity.
    """

    cls: str = "i64"
    width: int = 64


@dataclass
class U8(Typedecl):
    """
    A unsigned integer, 8 bits wide.

    The C emitter produces::

        uint8_t

    for the :class:`U8` entity.
    """

    cls: str = "u8"
    width: int = 8


@dataclass
class U16(Typedecl):
    """
    A unsigned integer, 16 bits wide".

    The C emitter produces::

        uint16_t

    for the :class:`U16` entity.
    """

    cls: str = "u16"
    width: int = 16


@dataclass
class U32(Typedecl):
    """
    An unsigned integer, 32 bits wide.

    The C emitter produces::

        uint32_t

    for the :class:`U32` entity.
    """

    cls: str = "u32"
    width: int = 32


@dataclass
class U64(Typedecl):
    """
    A unsigned integer, 64 bits wide.

    The C emitter produces::

        uint64_t

    for the :class:`U64` entity.
    """

    cls: str = "u64"
    width: int = 64


@dataclass
class F32(Typedecl):
    """
    A floating point numerical value, possibly 32 bits wide

    The C emitter produces::

        float

    For the :class:`F32` entity.
    """

    cls: str = "f32"
    width: int = 32


@dataclass
class F64(Typedecl):
    """
    A floating point numerical value, possibly 64 bits wide

    The C emitter produces::

        double

    For the :class:`F64` entity.
    """

    cls: str = "f64"
    width: int = 64


@dataclass
class Bool(Typedecl):
    """
    A boolean, at least 8 bits wide, equivalent to the C99 "_Bool" and available
    as "bool" as defined by the standardized "stdbool.h" header.

    The C emitter produces::

        bool

    for the :class:`Bool` entity.
    """

    cls: str = "bool"
    width: int = 8


@dataclass
class Char(Typedecl):
    """
    A character, at least 8 bits wide.

    The C emitter produces::

        char

    for the :class:`Char` entity.
    """

    cls: str = "char"
    width: int = 8


@dataclass
class String(Typedecl):
    """
    A string pointer

    The C emitter produces::

        const char *

    for the :class:`String` entity.
    """

    cls: str = "string"
    width: int = 8
    pointer: int = 1
    const: bool = True



#!/usr/bin/env python3
"""
The **yace** Interface Model (``yim``) is defined by plain-text files in YAML
format, these are described

* Yace Interface Model files (YAML Format)
* Entities
* Interface

The interface model was initially just a collection of YAML-files, globbed for
and merged together by the copiler using a YAML-parser. This left for two
things to be done:

* Documentation of the interface model
* Validation of the interface model

"""



#!/usr/bin/env python3
"""
C MACROS, one of biggest headaches when it comes to FFI interfaces, especially
for interpreted languages such as Python, since MACRO-expressions have no
symbols to "call" as these are all pre-preprocossed and gone thus, no longer
available.

Thus, **yace** only support the definition of MACRO values, such as::

    #define SOME_GLOBALLY_CONSTANT 0xACDC

Definitions such as these to translate fairly when into equivalent globally
available and constant constructs in other languages.
"""
import typing
from dataclasses import dataclass, field

from .entities import Entity
from .literals import LiteralDec, LiteralHex, LiteralString


@dataclass
class Define(Entity):
    """
    Representation of simple MACROs / definitions

    When emitting the C API, the Define holds information to generate things
    like::

        #define MAX_SOMETHING 128

    The other code-emitters will generate similar globally scoped definitions.
    """

    lit: typing.Union[LiteralHex, LiteralDec, LiteralString]

    cls: str = "define"



"""
The enumeration types consists of Enumerations (:class:`Enum`) and their values
:class:`EnumValue`. The values themselves are literals: :class:`LiteralDec`,
:class:`LiteralHex` and :class:`LiteralString`.
"""
import typing
from dataclasses import dataclass, field

from .entities import Entity
from .literals import LiteralDec, LiteralHex, LiteralString


@dataclass
class EnumValue(Entity):
    """
    Representation of values in :class:`.Enum`.

    By default, the C emitter produces::

        SYMBOL = lit.val

    for the :class:`EnumValue``.
    """

    lit: typing.Union[LiteralDec, LiteralHex, LiteralString] = None

    cls: str = "enum_value"


@dataclass
class Enum(Entity):
    """
    Representation of enumeration values, note that literals on the
    :class:`.EnumValue` are optional, an :class:`.Enum` with **non** literal
    :class:`.EnumValue` looks like::

        enum example {
            FOO,
            BAR,
            BAZ,
        };

    Whereas, an :class:`.Enum` with **literal** :class:`.EnumValue` looks
    like::

        enum example {
            FOO = 0xACDC,
            BAR = 0x1337,
            BAZ = 0xBEEF,
        };

    for the :class:`EnumValue``.
    """

    members: typing.List[EnumValue]

    cls: str = "enum"



#!/usr/bin/env python3
"""
Union Types behave the same as :class:`.structtypes.Struct`. The difference are
simply that they represent the declaration of a C Union rather than a C Struct.
"""
import typing
from dataclasses import dataclass, field

from .entities import Entity, Typedecl


@dataclass
class Union(Entity):
    """Representation of enumerations / collections of constants"""

    members: typing.List

    cls: str = "union"



#!/usr/bin/env python3
"""
Entities are the base classes representing the language constructs of C APIs in
the :class:`yace.model.interface.InterfaceModel`. The entities serve as
encapsulated type-checking of YAML-definitions of the language model.
"""
import typing
from dataclasses import dataclass, field

from .entities import Entity, Typedecl


@dataclass
class Bits(Entity):
    """Representation of a :class:`.Bitfield` member."""

    width: int

    cls: str = "bits"


@dataclass
class Bitfield(Entity):
    """Representation of enumerations / collections of constants"""

    members: typing.List[Bits]
    dtype: Typedecl

    cls: str = "bitfield"


@dataclass
class Field(Entity):
    """
    :class:`.Field` represents members of :class:`.Struct` and :class:`.Union`,
    utilized by compiler to emit code in the C API similar to::

        uint32_t bar;

    That is, within a ``struct/union`` block e.g.::

        struct foo {
            uint32_t bar;
        };

    The 'dtype' and 'width' members are utilized to produce a fitting
    type-declaration for the target language, as in the examples above for C.
    """

    dtype: Typedecl
    cls: str = "field"
    fmt: str = "%u"


@dataclass
class Struct(Entity):
    """Representation of enumerations / collections of constants"""

    members: typing.List

    cls: str = "struct"



#!/usr/bin/env python3
"""
Literals are at the heart of every language, with **yace** the following are
expressed:

* :class:`LiteralString`
* :class:`LiteralHex`
* :class:`LiteralDec`

These emit the usual suspects::

    #define PROJECT_DESCRIPTION "This is a literal string"

    enum jazz {
        FOO = 0xACDC,
        BAR = 0x1337,
        BAZ = 0xBEEF,
    };

    #define THRESHOLD 1337

The intent of these literals is just add a level of control to the values in
the entities :class:`.enumtypes.EnumValue` and :class:`entities.Define`.
"""
import typing
from dataclasses import dataclass, field


@dataclass
class LiteralString:
    """
    Representation of a an literal string e.g. in C::

        "this is a literal string"

    """

    val: str
    cls: str = "str"


@dataclass
class LiteralHex:
    """Representation of a an integer literal"""

    val: int
    cls: str = "hex"


@dataclass
class LiteralDec:
    """Representation of a an integer literal in decimal"""

    val: int
    cls: str = "dec"



from pathlib import Path

from setuptools import find_namespace_packages, setup

try:
    with open("README.rst", "r") as f:
        long_description = f.read()
except FileNotFoundError:
    long_description = ""

setup(
    name="yace",
    version="0.3.0",
    author="Simon A. F. Lund",
    author_email="os@safl.dk",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/safl/yace",
    license="BSD",
    setup_requires=["wheel"],
    install_requires=[
        "jinja2",
        "pyyaml",
        "setuptools>=60",
    ],
    python_requires=">=3.7",
    entry_points={"console_scripts": ["yace=yace.cli:main"]},
    packages=[
        "yace",
        "yace.model",
        "yace.templates",
        "yace.templates.c",
        "yace.templates.ctypes",
    ],
    package_dir={"": "src"},
    zip_safe=False,
    package_data={
        "": ["*.template"],
    },
    options={"bdist_wheel": {"universal": True}},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Framework :: Pytest",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development",
        "Topic :: Text Processing",
        "Topic :: Utilities",
    ],
)



"""
YAML ==> Interface(List[Entities]) ==> Emitter() ==> CodeTarget
"""
import typing
from pathlib import Path

from jinja2 import Environment, PackageLoader

from .model.entities import Entity
from .model.interface import InterfaceModel, Meta


class Emitter(object):
    """Produces code from :class:`yace.model.InterfaceModel`"""

    def __init__(
        self,
        model: InterfaceModel,
        meta: Meta,
        templates: typing.List[Path],
        output: Path,
    ):
        jenv = Environment(loader=PackageLoader("yace", "templates"))

        self.templates = {
            Path(f).stem: jenv.get_template(f)
            for f in jenv.list_templates()
            if f.endswith(".template")
        }
        self.model = model
        self.meta = meta
        self.output = output

    def emit_code(self, current: Entity, parent: typing.Optional[Entity]) -> str:
        """
        Recursively emit code for the 'current' model entity using the given
        'templates' for the given 'model'
        """

        content = []

        # Top-level "container" entities get a comment
        if not parent and current.cls in ["struct", "union", "enum"]:
            content.append(
                self.templates["comment"].render(
                    meta=self.meta, entity=current, parent=parent
                )
            )

        if current.cls in ["struct", "union"]:
            content.append(
                self.templates[f"{current.cls}_enter"].render(
                    meta=self.meta, entity=current, parent=parent
                )
            )
            for member in current.members:
                content += self.emit_code(member, current)
            content.append(
                self.templates[f"{current.cls}_exit"].render(
                    meta=self.meta, entity=current, parent=parent
                )
            )
        else:
            content.append(
                self.templates[current.cls].render(
                    meta=self.meta, entity=current, parent=parent
                )
            )

        return content

    def emit_api_def(self):
        """Emit header definitions"""

        content = []
        for entity in self.model.entities:
            content.append("\n".join(self.emit_code(entity, None)))

        with (self.output / f"lib{self.meta.prefix}.h").open("w") as hfile:
            hfile.write(
                self.templates["api_hdr"].render(
                    content="\n\n".join(content), meta=self.meta
                )
            )

    def emit_api_pp(self):
        """Emit pretty-printer functions"""

        in_and_out = [
            ("api_pp_hdr", f"lib{self.meta.prefix}_pp.h"),
            ("api_pp_src", f"{self.meta.prefix}_pp.c"),
        ]
        for template_name, fname in in_and_out:
            content = self.templates[template_name].render(
                meta=self.meta, entities=self.model.entities
            )
            with (self.output / fname).open("w") as hfile:
                hfile.write(content)

    def emit_api_test(self):
        """Emit test-program using definitions and pretty-printers"""

        in_and_out = [
            ("api_test_src", "test.c"),
        ]
        headers = [
            {"filename": "libnvme.h"},
            {"filename": "libnvme_pp.h"},
            {"filename": "libxnvme.h"},
            {"filename": "libxnvme_pp.h"},
        ]
        for template_name, fname in in_and_out:
            content = self.templates[template_name].render(
                meta=self.meta, entities=self.model.entities, headers=headers
            )
            with (self.output / fname).open("w") as hfile:
                hfile.write(content)

    def emit_docgen(self):
        """Emit test-program using definitions and pretty-printers"""

        headers = [
            {"filename": "libnvme.h"},
            {"filename": "libnvme_pp.h"},
            {"filename": "libxnvme.h"},
            {"filename": "libxnvme_pp.h"},
        ]
        content = self.templates["doxygen"].render(
            meta=self.meta, entities=self.model.entities, headers=headers
        )
        with (self.output / "doxy.cfg").open("w") as hfile:
            hfile.write(content)
